@@ -343,12 +343,72 @@ enum virgl_formats {
    VIRGL_FORMAT_A8L8_SNORM              = 260,
    VIRGL_FORMAT_A8L8_SRGB               = 261,
 
+   VIRGL_FORMAT_A1B5G5R5_UNORM          = 262,
+   VIRGL_FORMAT_A1R5G5B5_UNORM          = 263,
+   VIRGL_FORMAT_A2B10G10R10_UNORM       = 264,
+   VIRGL_FORMAT_A2R10G10B10_UNORM       = 265,
+   VIRGL_FORMAT_A4R4G4B4_UNORM          = 266,
+
    VIRGL_FORMAT_X8B8G8R8_SNORM          = 268,
 
+
+   /* etc2 compressed */
+   VIRGL_FORMAT_ETC2_RGB8               = 269,
+   VIRGL_FORMAT_ETC2_SRGB8              = 270,
+   VIRGL_FORMAT_ETC2_RGB8A1             = 271,
+   VIRGL_FORMAT_ETC2_SRGB8A1            = 272,
+   VIRGL_FORMAT_ETC2_RGBA8              = 273,
+   VIRGL_FORMAT_ETC2_SRGBA8             = 274,
+   VIRGL_FORMAT_ETC2_R11_UNORM          = 275,
+   VIRGL_FORMAT_ETC2_R11_SNORM          = 276,
+   VIRGL_FORMAT_ETC2_RG11_UNORM         = 277,
+   VIRGL_FORMAT_ETC2_RG11_SNORM         = 278,
+
+   VIRGL_FORMAT_ASTC_4x4                = 279,
+   VIRGL_FORMAT_ASTC_5x4                = 280,
+   VIRGL_FORMAT_ASTC_5x5                = 281,
+   VIRGL_FORMAT_ASTC_6x5                = 282,
+   VIRGL_FORMAT_ASTC_6x6                = 283,
+   VIRGL_FORMAT_ASTC_8x5                = 284,
+   VIRGL_FORMAT_ASTC_8x6                = 285,
+   VIRGL_FORMAT_ASTC_8x8                = 286,
+   VIRGL_FORMAT_ASTC_10x5               = 287,
+   VIRGL_FORMAT_ASTC_10x6               = 288,
+   VIRGL_FORMAT_ASTC_10x8               = 289,
+   VIRGL_FORMAT_ASTC_10x10              = 290,
+   VIRGL_FORMAT_ASTC_12x10              = 291,
+   VIRGL_FORMAT_ASTC_12x12              = 292,
+   VIRGL_FORMAT_ASTC_4x4_SRGB           = 293,
+   VIRGL_FORMAT_ASTC_5x4_SRGB           = 294,
+   VIRGL_FORMAT_ASTC_5x5_SRGB           = 295,
+   VIRGL_FORMAT_ASTC_6x5_SRGB           = 296,
+   VIRGL_FORMAT_ASTC_6x6_SRGB           = 297,
+   VIRGL_FORMAT_ASTC_8x5_SRGB           = 298,
+   VIRGL_FORMAT_ASTC_8x6_SRGB           = 299,
+   VIRGL_FORMAT_ASTC_8x8_SRGB           = 300,
+   VIRGL_FORMAT_ASTC_10x5_SRGB          = 301,
+   VIRGL_FORMAT_ASTC_10x6_SRGB          = 302,
+   VIRGL_FORMAT_ASTC_10x8_SRGB          = 303,
+   VIRGL_FORMAT_ASTC_10x10_SRGB         = 304,
+   VIRGL_FORMAT_ASTC_12x10_SRGB         = 305,
+   VIRGL_FORMAT_ASTC_12x12_SRGB         = 306,
+
    VIRGL_FORMAT_R10G10B10X2_UNORM       = 308,
    VIRGL_FORMAT_A4B4G4R4_UNORM          = 311,
 
    VIRGL_FORMAT_R8_SRGB                 = 312,
+   VIRGL_FORMAT_R8G8_SRGB               = 313,
+
+   VIRGL_FORMAT_P010                    = 314,
+   VIRGL_FORMAT_P012                    = 315,
+   VIRGL_FORMAT_P016                    = 316,
+
+   VIRGL_FORMAT_B8G8R8_UNORM            = 317,
+   VIRGL_FORMAT_R3G3B2_UNORM            = 318,
+   VIRGL_FORMAT_R4G4B4A4_UNORM          = 319,
+   VIRGL_FORMAT_R5G5B5A1_UNORM          = 320,
+   VIRGL_FORMAT_R5G6B5_UNORM            = 321,
+
    VIRGL_FORMAT_MAX /* = PIPE_FORMAT_COUNT */,
 
    /* Below formats must not be used in the guest. */
@@ -388,7 +448,21 @@ enum virgl_formats {
 #define VIRGL_CAP_COPY_TRANSFER        (1 << 26)
 #define VIRGL_CAP_CLIP_HALFZ           (1 << 27)
 #define VIRGL_CAP_APP_TWEAK_SUPPORT    (1 << 28)
-#define VIRGL_CAP_BGRA_SRGB_IS_EMULATED  (1 << 29)
+#define VIRGL_CAP_BGRA_SRGB_IS_EMULATED (1 << 29)
+#define VIRGL_CAP_CLEAR_TEXTURE        (1 << 30)
+#define VIRGL_CAP_ARB_BUFFER_STORAGE   (1 << 31)
+
+/* These are used by the capability_bits_v2 field in virgl_caps_v2. */
+#define VIRGL_CAP_V2_BLEND_EQUATION       (1 << 0)
+#define VIRGL_CAP_V2_UNTYPED_RESOURCE     (1 << 1)
+#define VIRGL_CAP_V2_VIDEO_MEMORY         (1 << 2)
+#define VIRGL_CAP_V2_MEMINFO              (1 << 3)
+#define VIRGL_CAP_V2_STRING_MARKER        (1 << 4)
+#define VIRGL_CAP_V2_DIFFERENT_GPU        (1 << 5)
+#define VIRGL_CAP_V2_IMPLICIT_MSAA        (1 << 6)
+#define VIRGL_CAP_V2_COPY_TRANSFER_BOTH_DIRECTIONS (1 << 7)
+#define VIRGL_CAP_V2_SCANOUT_USES_GBM     (1 << 8)
+#define VIRGL_CAP_V2_SSO                  (1 << 9)
 
 /* virgl bind flags - these are compatible with mesa 10.5 gallium.
  * but are fixed, no other should be passed to virgl either.
@@ -413,10 +487,22 @@ enum virgl_formats {
 #define VIRGL_BIND_STAGING       (1 << 19)
 #define VIRGL_BIND_SHARED        (1 << 20)
 
-#define VIRGL_BIND_PREFER_EMULATED_BGRA  (1 << 21)
+#define VIRGL_BIND_PREFER_EMULATED_BGRA  (1 << 21) /* non-functional */
 
 #define VIRGL_BIND_LINEAR (1 << 22)
 
+#define VIRGL_BIND_SHARED_SUBFLAGS (0xff << 24)
+
+#define VIRGL_BIND_MINIGBM_CAMERA_WRITE (1 << 24)
+#define VIRGL_BIND_MINIGBM_CAMERA_READ (1 << 25)
+#define VIRGL_BIND_MINIGBM_HW_VIDEO_DECODER (1 << 26)
+#define VIRGL_BIND_MINIGBM_HW_VIDEO_ENCODER (1 << 27)
+#define VIRGL_BIND_MINIGBM_SW_READ_OFTEN (1 << 28)
+#define VIRGL_BIND_MINIGBM_SW_READ_RARELY (1 << 29)
+#define VIRGL_BIND_MINIGBM_SW_WRITE_OFTEN (1 << 30)
+#define VIRGL_BIND_MINIGBM_SW_WRITE_RARELY (1 << 31)
+#define VIRGL_BIND_MINIGBM_PROTECTED (0xf << 28) // Mutually exclusive with SW_ flags
+
 struct virgl_caps_bool_set1 {
         unsigned indep_blend_enable:1;
         unsigned indep_blend_func:1;
@@ -478,6 +564,26 @@ struct virgl_caps_v1 {
         uint32_t max_texture_gather_components;
 };
 
+struct virgl_video_caps {
+        uint32_t profile:8;
+        uint32_t entrypoint:8;
+        uint32_t max_level:8;
+        uint32_t stacked_frames:8;
+
+        uint32_t max_width:16;
+        uint32_t max_height:16;
+
+        uint32_t prefered_format:16;
+        uint32_t max_macroblocks:16;
+
+        uint32_t npot_texture:1;
+        uint32_t supports_progressive:1;
+        uint32_t supports_interlaced:1;
+        uint32_t prefers_interlaced:1;
+        uint32_t max_temporal_layers:8;
+        uint32_t reserved:20;
+};
+
 /*
  * This struct should be growable when used in capset 2,
  * so we shouldn't have to add a v3 ever.
@@ -528,6 +634,16 @@ struct virgl_caps_v2 {
         uint32_t host_feature_check_version;
         struct virgl_supported_format_mask supported_readback_formats;
         struct virgl_supported_format_mask scanout;
+        uint32_t capability_bits_v2;
+        uint32_t max_video_memory;
+        char renderer[64];
+        float max_anisotropy;
+        uint32_t max_texture_image_units;
+        struct virgl_supported_format_mask supported_multisample_formats;
+        uint32_t max_const_buffer_size[6]; // PIPE_SHADER_TYPES
+        uint32_t num_video_caps;
+        struct virgl_video_caps video_caps[32];
+        uint32_t max_uniform_block_size;
 };
 
 union virgl_caps {
@@ -556,8 +672,16 @@ enum virgl_ctx_errors {
         VIRGL_ERROR_CTX_ILLEGAL_FORMAT,
         VIRGL_ERROR_CTX_ILLEGAL_SAMPLER_VIEW_TARGET,
         VIRGL_ERROR_CTX_TRANSFER_IOV_BOUNDS,
-        VIRGL_ERROR_CTX_ILLEGAL_DUAL_SRC_BLEND
+        VIRGL_ERROR_CTX_ILLEGAL_DUAL_SRC_BLEND,
+        VIRGL_ERROR_CTX_UNSUPPORTED_FUNCTION,
+        VIRGL_ERROR_CTX_ILLEGAL_PROGRAM_PIPELINE,
 };
 
+/**
+ * Flags for the driver about resource behaviour:
+ */
 #define VIRGL_RESOURCE_Y_0_TOP (1 << 0)
+#define VIRGL_RESOURCE_FLAG_MAP_PERSISTENT (1 << 1)
+#define VIRGL_RESOURCE_FLAG_MAP_COHERENT   (1 << 2)
+
 #endif
